@@ -2,12 +2,12 @@ package RedisDB;
 
 use strict;
 use warnings;
-our $VERSION = "2.30";
+our $VERSION = "2.32";
 $VERSION = eval $VERSION;
 
 use RedisDB::Error;
 use RedisDB::Parser;
-use IO::Socket::INET;
+use IO::Socket::IP;
 use IO::Socket::UNIX;
 use Socket qw(MSG_DONTWAIT MSG_NOSIGNAL SO_RCVTIMEO SO_SNDTIMEO);
 use POSIX qw(:errno_h);
@@ -272,7 +272,7 @@ sub _connect {
             my $delay;
             while ( not $self->{_socket} and $attempts ) {
                 sleep $delay if $delay;
-                $self->{_socket} = IO::Socket::INET->new(
+                $self->{_socket} = IO::Socket::IP->new(
                     PeerAddr => $self->{host},
                     PeerPort => $self->{port},
                     Proto    => 'tcp',
@@ -695,6 +695,25 @@ sub get_reply {
         croak $res;
     }
 
+    if ( $self->{_subscription_loop} ) {
+        confess "Expected multi-bulk reply, but got $res" unless ref $res;
+        if ( $res->[0] eq 'message' ) {
+            $self->{_subscribed}{ $res->[1] }( $self, $res->[1], undef, $res->[2] )
+              if $self->{_subscribed}{ $res->[1] };
+        }
+        elsif ( $res->[0] eq 'pmessage' ) {
+            $self->{_psubscribed}{ $res->[1] }( $self, $res->[2], $res->[1], $res->[3] )
+              if $self->{_psubscribed}{ $res->[1] };
+        }
+        elsif ( $res->[0] =~ /^p?(un)?subscribe/ ) {
+
+            # ignore
+        }
+        else {
+            confess "Got unknown reply $res->[0] in subscription mode";
+        }
+    }
+
     return $res;
 }
 
@@ -784,7 +803,7 @@ my @commands = qw(
   hincrby	hincrbyfloat	hkeys	hlen	hmget	hscan	hmset	hset	hsetnx	hvals	incr	incrby
   incrbyfloat	keys	lastsave	lindex	linsert	llen	lpop	lpush	lpushx
   lrange	lrem	lset	ltrim	mget	migrate	move	mset	msetnx	object	object_refcount
-  object_encoding	object_idletime	persist	pexpire	pexpireat	ping	psetex	pttl
+  object_encoding	object_idletime	persist	pexpire	pexpireat	pfadd	pfcount	pfmerge	ping	psetex	pttl
   pubsub	pubsub_channels	pubsub_numsub	pubsub_numpat
   publish	quit	randomkey	rename	renamenx	restore	rpop	rpoplpush
   rpush	rpushx	sadd	save	scan	scard	script	script_exists   script_flush    script_kill
@@ -792,8 +811,8 @@ my @commands = qw(
   setbit	setex	setnx	setrange	sinter	sinterstore
   sismember	slaveof	slowlog smembers	smove	sort	spop	srandmember
   srem	sscan	strlen	sunion	sunionstore	sync	time    ttl	type	unwatch watch
-  zadd	zcard
-  zcount	zincrby	zinterstore	zrange	zrangebyscore	zrank	zrem
+  zadd	zcard	zcount	zincrby	zinterstore	zlexcount	zrange	zrangebylex
+  zrangebyscore	zrank	zrem	zremrangebylex
   zremrangebyrank   zremrangebyscore	zrevrange	zrevrangebyscore	zrevrank
   zscan	zscore	zunionstore
 );
@@ -822,22 +841,23 @@ I<execute>, waits for the reply from the server, and returns it. E.g.:
     $redis->send_command("get", $key, sub { $val = $_[1] });
 
 The following wrapper methods are defined: append, auth, bgrewriteaof, bgsave,
-bitcount, bitop, bitpos, blpop, brpop, brpoplpush, client, client_kill, client_getname,
-client_setname, config, config_get, config_set, config_resetstat, config_rewrite, dbsize,
-debug_object, debug_segfault, decr, decrby, del, dump, echo, eval, evalsha,
-exists, expire, expireat, flushall, flushdb, get, getbit, getrange, getset,
-hdel, hexists, hget, hgetall, hincrby, hincrbyfloat, hkeys, hlen, hmget, hscan, hmset,
-hset, hsetnx, hvals, incr, incrby, incrbyfloat, keys, lastsave, lindex,
-linsert, llen, lpop, lpush, lpushx, lrange, lrem, lset, ltrim, mget, migrate,
-move, mset, msetnx, object, object_refcount, object_encoding, object_idletime,
-persist, pexpire, pexpireat, ping, psetex, pttl, publish,
-pubsub, pubsub_channels, pubsub_numsub, pubsub_numpat, quit, randomkey,
-rename, renamenx, restore, rpop, rpoplpush, rpush, rpushx, sadd, save, scan, scard,
-script, script_exists, script_flush, script_kill, script_load, sdiff,
-sdiffstore, select, set, setbit, setex, setnx, setrange, sinter, sinterstore,
-sismember, slaveof, slowlog, smembers, smove, sort, spop, srandmember, srem,
-sscan strlen, sunion, sunionstore, sync, time, ttl, type, unwatch, watch, zadd,
-zcard, zcount, zincrby, zinterstore, zrange, zrangebyscore, zrank, zrem,
+bitcount, bitop, bitpos, blpop, brpop, brpoplpush, client, client_kill,
+client_getname, client_setname, config, config_get, config_set,
+config_resetstat, config_rewrite, dbsize, debug_object, debug_segfault, decr,
+decrby, del, dump, echo, eval, evalsha, exists, expire, expireat, flushall,
+flushdb, get, getbit, getrange, getset, hdel, hexists, hget, hgetall, hincrby,
+hincrbyfloat, hkeys, hlen, hmget, hscan, hmset, hset, hsetnx, hvals, incr,
+incrby, incrbyfloat, keys, lastsave, lindex, linsert, llen, lpop, lpush,
+lpushx, lrange, lrem, lset, ltrim, mget, migrate, move, mset, msetnx, object,
+object_refcount, object_encoding, object_idletime, persist, pexpire, pexpireat,
+pfadd, pfcount, pfmerge, ping, psetex, pttl, publish, pubsub, pubsub_channels, pubsub_numsub,
+pubsub_numpat, quit, randomkey, rename, renamenx, restore, rpop, rpoplpush,
+rpush, rpushx, sadd, save, scan, scard, script, script_exists, script_flush,
+script_kill, script_load, sdiff, sdiffstore, select, set, setbit, setex, setnx,
+setrange, sinter, sinterstore, sismember, slaveof, slowlog, smembers, smove,
+sort, spop, srandmember, srem, sscan strlen, sunion, sunionstore, sync, time,
+ttl, type, unwatch, watch, zadd, zcard, zcount, zincrby, zinterstore,
+zlexcount, zrange, zrangebylex, zrangebyscore, zrank, zrem, zremrangebylex,
 zremrangebyrank, zremrangebyscore, zrevrange, zrevrangebyscore, zrevrank,
 zscan, zscore, zunionstore.
 
@@ -1140,43 +1160,70 @@ or using L</"WRAPPER METHODS"> you can rewrite it as:
 
 RedisDB supports subscriptions to redis channels. In the subscription mode you
 can subscribe to some channels and receive all the messages sent to these
-channels.  Every time RedisDB receives a message for the channel it invokes a
-callback provided by the user. User can specify different callbacks for the
-different channels.  When in the subscription mode you can subscribe to
-additional channels, or unsubscribe from the channels you subscribed to, but
-you can't use any other redis commands like set, get, etc. Here is the example
-of running in the subscription mode:
+channels. You can subscribe to channels and then manually check messages using
+I<get_reply> method, or you can invoke I<subscription_loop> method, which will
+block in loop waiting for messages and invoking callback for each received
+message.  In the first case you can use I<subscribe> and I<psubscribe> methods
+to subscribe to channels and then you can use I<get_reply> method to get
+messages from the channel:
+
+    $redis->subscribe(
+        foo => sub {
+            my ( $redis, $channel, $patern, $message ) = @_;
+            print "Foo: $message\n";
+        }
+    );
+    # Wait for messages
+    $res = $redis->get_reply;
+
+I<get_reply> method for messages from the channel will invoke callback
+specified as the second optional argument of the I<subscribe> method and will
+also return raw replies from the server, both for messages from the channels
+and for informational messages from the redis server. If you do not want to
+block in I<get_reply> method, you can check if there are any messages using
+I<reply_ready> method.
+
+In the second case you invoke I<subscription_loop> method, it subscribes to
+specified channels and waits for messages, when a message arrived it invokes
+callback defined for the channel from which the message came. Here is an
+example:
 
     my $message_cb = sub {
-        my ($redis, $channel, $pattern, $message) = @_;
+        my ( $redis, $channel, $pattern, $message ) = @_;
         print "$channel: $message\n";
     };
-    
+
     my $control_cb = sub {
-        my ($redis, $channel, $pattern, $message) = @_;
-        if ($channel eq 'control.quit') {
+        my ( $redis, $channel, $pattern, $message ) = @_;
+        if ( $channel eq 'control.quit' ) {
             $redis->unsubscribe;
             $redis->punsubscribe;
         }
-        elsif ($channel eq 'control.subscribe') {
+        elsif ( $channel eq 'control.subscribe' ) {
             $redis->subscribe($message);
         }
     };
-    
-    subscription_loop(
-        subscribe => [ 'news',  ],
-        psubscribe => [ 'control.*' => $control_cb ],
+
+    $redis->subscription_loop(
+        subscribe        => [ 'news', ],
+        psubscribe       => [ 'control.*' => $control_cb ],
         default_callback => $message_cb,
     );
 
-subscription_loop will subscribe you to the news channel and control.*
+subscription_loop will subscribe you to the "news" channel and "control.*"
 channels. It will call specified callbacks every time a new message received.
-You can subscribe to additional channels sending their names to the
-control.subscribe channel. You can unsubscribe from all the channels by sending
-a message to the control.quit channel. Every callback receives four arguments:
-the RedisDB object, the channel for which the message was received, the pattern
-if you subscribed to this channel using I<psubscribe> method, and the message
-itself.
+When message came from "control.subscribe" channel, callback subscribes to an
+additional channel. When message came from "control.quit" channel, callback
+unsubscribes from all channels.
+
+Callbacks used in subscription mode receive four arguments: the RedisDB object,
+the channel from which the message came, the pattern if you subscribed to this
+channel using I<psubscribe> method, and the message itself.
+
+Once you switched into subscription mode using either I<subscribe> or
+I<psubscribe> command, or by entering I<subscription_loop>, you only can send
+I<subscribe>, I<psubscribe>, I<unsubscribe>, and I<punsubscribe> commands to
+the server, other commands will throw an exception.
 
 You can publish messages into the channels using the I<publish> method. This
 method should be called when you in the normal mode, and can't be used while
@@ -1225,11 +1272,11 @@ for every channel you are going to subscribe.
 
 sub subscription_loop {
     my ( $self, %args ) = @_;
-    croak "Already in subscription loop" if $self->{_subscription_loop};
+    croak "Already in subscription loop" if $self->{_subscription_loop} > 0;
     croak "You can't start subscription loop while in pipelining mode."
       if $self->replies_to_fetch;
-    $self->{_subscribed}        = {};
-    $self->{_psubscribed}       = {};
+    $self->{_subscribed}  ||= {};
+    $self->{_psubscribed} ||= {};
     $self->{_subscription_cb}   = $args{default_callback};
     $self->{_subscription_loop} = 1;
     $self->{_parser}->set_default_callback( \&_queue );
@@ -1252,23 +1299,7 @@ sub subscription_loop {
       unless ( keys %{ $self->{_subscribed} } or keys %{ $self->{_psubscribed} } );
 
     while ( $self->{_subscription_loop} ) {
-        my $msg = $self->get_reply;
-        confess "Expected multi-bulk reply, but got $msg" unless ref $msg;
-        if ( $msg->[0] eq 'message' ) {
-            $self->{_subscribed}{ $msg->[1] }( $self, $msg->[1], undef, $msg->[2] )
-              if $self->{_subscribed}{ $msg->[1] };
-        }
-        elsif ( $msg->[0] eq 'pmessage' ) {
-            $self->{_psubscribed}{ $msg->[1] }( $self, $msg->[2], $msg->[1], $msg->[3] )
-              if $self->{_psubscribed}{ $msg->[1] };
-        }
-        elsif ( $msg->[0] =~ /^p?(un)?subscribe/ ) {
-
-            # ignore
-        }
-        else {
-            confess "Got unknown reply $msg->[0] in subscription mode";
-        }
+        $self->get_reply;
     }
     return;
 }
@@ -1276,8 +1307,8 @@ sub subscription_loop {
 =head2 $self->subscribe($channel[, $callback])
 
 Subscribe to the I<$channel>. If I<$callback> is not specified, default
-callback will be used. If you are invoking I<subscribe> outside of subscription
-loop, I<$callback> is ignored.
+callback will be used in subscription loop, or messages will be returned by
+I<get_reply> if you are not using subscription loop.
 
 =cut
 
@@ -1294,18 +1325,18 @@ sub subscribe {
           or croak "Callback for $channel not specified, neither default callback defined";
     }
     else {
-        $callback = 1;
+        $callback ||= sub { 1 };
     }
     $self->{_subscribed}{$channel} = $callback;
-    $self->send_command( "SUBSCRIBE", $channel );
+    $self->send_command( "SUBSCRIBE", $channel, \&_queue );
     return;
 }
 
 =head2 $self->psubscribe($pattern[, $callback])
 
 Subscribe to channels matching I<$pattern>. If I<$callback> is not specified,
-default callback will be used. If you are invoking I<psubscribe> outside of
-subscription loop, I<$callback> is ignored.
+default callback will be used in subscription loop, or messages will be
+returned by I<get_reply> if you are not using subscription loop.
 
 =cut
 
@@ -1322,10 +1353,10 @@ sub psubscribe {
           or croak "Callback for $channel not specified, neither default callback defined";
     }
     else {
-        $callback = 1;
+        $callback ||= sub { 1 };
     }
     $self->{_psubscribed}{$channel} = $callback;
-    $self->send_command( "PSUBSCRIBE", $channel );
+    $self->send_command( "PSUBSCRIBE", $channel, \&_queue );
     return;
 }
 
